/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


#include "mavlib_collide.h"
#include <stdio.h>
#include <math.h>

/***************************************************************/
/* main collision library routines and initialization function */
/***************************************************************/

int mavlib_collideInitialized= MAV_FALSE;

int mav_collidePolygonObjectPointGet (MAV_polygon *poly, MAV_object *obj, MAV_vector *hit_pt)
{
  int vertex;
  MAV_vector next, curr_norm, edge, plane_point[MAV_COLLIDE_MAX_POINTS];
  float length;
  MAV_BB bb;
  int in_count[MAV_COLLIDE_MAX_POINTS], count, pts;
  MAV_line ln;
  MAV_objectIntersection oi;
  MAV_line plane_ln;

/* see if obj intersects plane of poly */
  plane_ln.pt= poly->vert[0];
  plane_ln.dir= poly->norm;

  pts= mav_callbackPlanePointExec (mav_win_current, obj, plane_ln, plane_point);

/* plane_point contains array of points on the plane of poly */
/* that intersect obj */

/* test if any of these points are inside poly and test each edge */
/* of poly against obj */

  for (count= 0; count < pts; count ++)
    in_count[count]= 0;

  for (vertex= 0; vertex < poly->np; vertex ++) {
    ln.pt= poly->vert[vertex];
    next= poly->vert[(vertex+1)%(poly->np)];

/* get current edge vector */
    edge= mav_vectorSub (next, ln.pt);
    length= mav_vectorDotProduct (edge, edge);
    if (length>0) length= sqrt (length);
    edge= mav_vectorNormalize (edge);

/* test this edge against plane_point array */
    curr_norm= mav_vectorCrossProduct (poly->norm, edge);
    for (count= 0; count < pts; count ++)
      if (mav_vectorDotProduct (plane_point[count], curr_norm) >=
	mav_vectorDotProduct (ln.pt, curr_norm))
		in_count[count] ++;

/* intersect this edge with obj */
    ln.dir= edge;

    if (mav_callbackQuery (mav_callback_planePoint, mav_win_current, obj)) {
      if (mav_callbackIntersectExec (mav_win_current, obj, ln, &oi))
	if ((oi.pt1 > -MAV_EPSILON) && (oi.pt1 <= length)) {
	  hit_pt->x= ln.pt.x + ln.dir.x * oi.pt1;
	  hit_pt->y= ln.pt.y + ln.dir.y * oi.pt1;
	  hit_pt->z= ln.pt.z + ln.dir.z * oi.pt1;
	  return 1;
	}
    } else {
      mav_callbackBBExec (mav_win_current, obj, &bb);
      if (mav_BBIntersect (bb, ln, &oi))
	if ((oi.pt1 > -MAV_EPSILON) && (oi.pt1 <= length)) {
          hit_pt->x= ln.pt.x + ln.dir.x * oi.pt1;
          hit_pt->y= ln.pt.y + ln.dir.y * oi.pt1;
          hit_pt->z= ln.pt.z + ln.dir.z * oi.pt1;
	  return 1;
	}
    }
  }

/* now check the results of the edge tests */
  for (count= 0; count < pts; count ++)
    if ((in_count[count]==poly->np)||(in_count[count]==0)) {
      *hit_pt= plane_point[count];
      return 1;
    }

/* got here then no intersection */
  return 0;
}


int mav_collidePolygonObjectTest (MAV_polygon *poly, MAV_object *obj)
{
  MAV_vector dummy;

  return mav_collidePolygonObjectPointGet (poly, obj, &dummy);
}

int mav_collidePolygonObjectVectorPointGet (MAV_polygon *poly, MAV_object *obj, MAV_vector dir, float dist, MAV_vector *hit_pt)
{
  MAV_polygon test_poly;
  int count;
  MAV_vector scaled_dir;
  MAV_line test_ln;
  MAV_objectIntersection oi;

  scaled_dir= mav_vectorScalar (dir, dist);

  test_poly.vert= (MAV_vector *) mav_malloc (poly->np * sizeof (MAV_vector));

  test_poly.np= poly->np;

  test_poly.matrix= MAV_ID_MATRIX;

/* test far side */
  for (count= 0; count < poly->np; count ++)
    test_poly.vert[count]= mav_vectorAdd (poly->vert[count], scaled_dir);

  test_poly.norm= mav_vectorCrossProduct (mav_vectorSub (test_poly.vert[1],
		test_poly.vert[0]),
		mav_vectorSub (test_poly.vert[2], test_poly.vert[1]));
  test_poly.norm= mav_vectorNormalize (test_poly.norm);

  if (mav_collidePolygonObjectPointGet (&test_poly, obj, hit_pt)) {
    mav_free (test_poly.vert);
    return 1;
  }

/* test near side */

  poly->norm= test_poly.norm;

  if (mav_collidePolygonObjectPointGet (poly, obj, hit_pt)) {
    mav_free (test_poly.vert);
    return 1;
  }


/* get a point on obj */
  mav_callbackGetPointExec (mav_win_current, obj, &(test_ln.pt));
  test_ln.dir= dir;

/* test a ray from this point (only need to check end polygon) */
  if (mav_linePolygonIntersection (&test_poly, test_ln, &oi))
    if ((oi.pt1 > -MAV_EPSILON) && (oi.pt1 < dist)) {
      *hit_pt= mav_vectorAdd (test_ln.pt,
		mav_vectorScalar (test_ln.dir, oi.pt1));
      mav_free (test_poly.vert);
      return 1;
    }

  test_poly.np= 4;

/* test each side of polygon volume */
  for (count= 0; count < poly->np; count ++) {
    test_poly.vert[0]= poly->vert[count];
    test_poly.vert[1]= poly->vert[(count+1)%(poly->np)];
    test_poly.vert[2]= mav_vectorAdd (test_poly.vert[1], scaled_dir);
    test_poly.vert[3]= mav_vectorAdd (test_poly.vert[0], scaled_dir);

    test_poly.norm= mav_vectorCrossProduct (mav_vectorSub (test_poly.vert[1],
                test_poly.vert[0]),
                mav_vectorSub (test_poly.vert[2], test_poly.vert[1]));
    test_poly.norm= mav_vectorNormalize (test_poly.norm);

    if (mav_collidePolygonObjectPointGet (&test_poly, obj, hit_pt)) {
      mav_free (test_poly.vert);
      return 1;
    }
  }

  mav_free (test_poly.vert);
  return 0;
}

int mav_collidePolygonObjectVectorTest (MAV_polygon *poly, MAV_object *obj, MAV_vector dir, float dist)
{
  MAV_vector dummy;

  return mav_collidePolygonObjectVectorPointGet (poly, obj, dir, dist, &dummy);
}


/* variables for mavlib_culledPolygonTest to use */
int mavlib_collideCollision;
MAV_vector mavlib_collideDir;
float mavlib_collideDist;
MAV_polygon *mavlib_collidePoly;

void mavlib_culledPolygonTest (MAV_object *obj, MAV_drawInfo *di, void *params)
{
  if (!mavlib_collideCollision)
    if (mav_collidePolygonObjectVectorTest (mavlib_collidePoly, obj,
	mavlib_collideDir, mavlib_collideDist))
      mavlib_collideCollision= 1;
}

int mav_collidePolygonCullVectorTest (MAV_polygon *poly, MAV_vector dir, float dist, MAV_list *sms_list)
{
/* NOTE - assumes a 4 sided polygon */
  MAV_SMS *sms;
  MAV_drawInfo di;
  MAV_SMSExecFn cull_fn;
  MAV_vector pt;
  int rv= 0;

/* set variables for mavlib_culledPolygonTest */
  mavlib_collidePoly= poly;
  mavlib_collideDir= dir;
  mavlib_collideDist= dist;
  mavlib_collideCollision= 0;

/* compute clip planes for the region poly will move through */
  di.cp.planes[0].norm= mav_vectorCrossProduct (
	mav_vectorSub (poly->vert[1], poly->vert[0]),
	mav_vectorSub (poly->vert[2], poly->vert[1]));

/* get a point on opposite side to poly */
  pt= mav_vectorAdd (poly->vert[0], mav_vectorScalar (dir, dist));

/* set clip plane normals (depends on normal direction) */
  if (mav_vectorDotProduct (di.cp.planes[0].norm, dir)<0.0) {
    di.cp.planes[1].norm= di.cp.planes[0].norm;
    di.cp.planes[0].norm= mav_vectorScalar (di.cp.planes[0].norm, -1.0);
    di.cp.planes[2].norm= mav_vectorCrossProduct (
        dir, mav_vectorSub (poly->vert[1], poly->vert[0]));
    di.cp.planes[3].norm= mav_vectorCrossProduct (
        dir, mav_vectorSub (poly->vert[3], poly->vert[2]));
    di.cp.planes[4].norm= mav_vectorCrossProduct (
        dir, mav_vectorSub (poly->vert[2], poly->vert[1]));
    di.cp.planes[5].norm= mav_vectorCrossProduct (
        dir, mav_vectorSub (poly->vert[0], poly->vert[3]));
  } else {
    di.cp.planes[1].norm= mav_vectorScalar (di.cp.planes[0].norm, -1.0);
    di.cp.planes[2].norm= mav_vectorCrossProduct (
        mav_vectorSub (poly->vert[1], poly->vert[0]), dir);
    di.cp.planes[3].norm= mav_vectorCrossProduct (
        mav_vectorSub (poly->vert[3], poly->vert[2]), dir);
    di.cp.planes[4].norm= mav_vectorCrossProduct (
        mav_vectorSub (poly->vert[2], poly->vert[1]), dir);
    di.cp.planes[5].norm= mav_vectorCrossProduct (
        mav_vectorSub (poly->vert[0], poly->vert[3]), dir);
  }

  di.cp.planes[0].norm= mav_vectorNormalize (di.cp.planes[0].norm);
  di.cp.planes[1].norm= mav_vectorNormalize (di.cp.planes[1].norm);
  di.cp.planes[2].norm= mav_vectorNormalize (di.cp.planes[2].norm);
  di.cp.planes[3].norm= mav_vectorNormalize (di.cp.planes[3].norm);
  di.cp.planes[4].norm= mav_vectorNormalize (di.cp.planes[4].norm);
  di.cp.planes[5].norm= mav_vectorNormalize (di.cp.planes[5].norm);

  di.cp.planes[0].d= mav_vectorDotProduct (di.cp.planes[0].norm, pt);
  di.cp.planes[1].d= mav_vectorDotProduct (di.cp.planes[1].norm,
        poly->vert[0]);
  di.cp.planes[2].d= mav_vectorDotProduct (di.cp.planes[2].norm,
        poly->vert[0]);
  di.cp.planes[3].d= mav_vectorDotProduct (di.cp.planes[3].norm,
        poly->vert[2]);
  di.cp.planes[4].d= mav_vectorDotProduct (di.cp.planes[4].norm,
        poly->vert[1]);
  di.cp.planes[5].d= mav_vectorDotProduct (di.cp.planes[5].norm,
        poly->vert[0]);

  di.cp.num= 6;

/* set SMSExecFn.fn */
  cull_fn.fn= mavlib_culledPolygonTest;

/* step through sms_list culling each SMS */
  mav_listPointerReset (sms_list);
  while (mav_listItemNext (sms_list, (void *) &sms))
    mav_SMSCallbackExecFnExec (sms, &di, &cull_fn);

/* check result of collision tests */
  if (!mavlib_collideCollision) rv= 1;

  return rv;
}

/* Routines to initialise the module */

char *mav_collideModuleID (void)
{
  return "Collision detection";
}

void mav_planePointInit (void);

void mav_collideModuleInit (void) {
  if (!mavlib_collideInitialized) {
    mavlib_getPointInit ();
    mav_planePointInit ();
    mavlib_collideInitialized= MAV_TRUE;
  }
}

